#
#   irrigation_sim.py  -  irrigation simulator
#

# Standard Library
from math import *

# Additional Packages
from Numeric import *
from MLab import *
import pylab
import Image

# User-defined Packages
import irrigation_tools as tools

class Sprinkler:

    distance_from_root = 0.0 # distance from the root point.


    # position vector of the root point.
    x = 0.0
    y = 0.0
    phi = 0.0
    arc = [1]  # water a half circle
    # arc = [1,-1]
    # arc = [-1]

    # constants
    nozzle_diameter = 0.6 # cm
    percent_water = 0.9 # percentage of water retained

    # derived values
    nozzle_flow = 0.0
    nozzle_velocity = 0.0
    dispersion_radius = 0.0
    dispersion_area = 0.0

    def __init__(self, nozzle_flow, theta):
        self.nozzle_flow = nozzle_flow * tools.in_conv["liters_per_minute"]
        self.nozzle_velocity = tools.velocity(self.nozzle_flow, 
         self.nozzle_diameter,
         in_units=["cubic_meters_per_second", "centimeters"])
        self.dispersion_radius = tools.radius(self.nozzle_velocity, theta)
        self.dispersion_area = tools.area(self.dispersion_radius)

    # water depth as a function of distance from the nozzle.
    @tools.units
    def dispersion_profile(self, distance_from_nozzle):
        return 3.0*(-distance_from_nozzle/self.dispersion_radius + \
         self.percent_water)*(self.nozzle_flow/self.dispersion_area)

    @tools.units
    def dispersion(self, dc):
        return self.dispersion_profile(dc)

    def move(self, x, y, phi):
        self.x = x
        self.y = y
        self.phi = phi

    def coords(self):
        return (self.x + self.distance_from_root * cos(self.phi),
         self.y + self.distance_from_root * sin(self.phi))

class SprinklerSystem:

    inflow = 150.0 # liters per minute
    x = 0.0 # coords of pipe end
    y = 0.0 
    phi = pi/2 # angle of the pipe

    def __init__(self, sprinkler_positions, theta):
        self.sprinklers = []
        n = len(sprinkler_positions)
        for ii in range(n):
            s = Sprinkler(self.inflow/float(n), theta)
            s.distance_from_root = sprinkler_positions[ii]
            s.x = self.x
            s.y = self.y
            s.phi = self.phi
            self.sprinklers.append(s)

    def run_sprinklers(self):
        for sprinkler in self.sprinklers:
            yield sprinkler

    def move(self,x,y,phi):
        self.x = x
        self.y = y
        self.phi = phi
        for sprinkler in self.sprinklers:
            sprinkler.move(x,y,phi)
        


class Field:

    max_hourly_water = 0.75 # cm
    width = 80.0
    height = 30.0
    sectors = array([])

    def __init__(self, resolution, max_hourly_water):
        self.resolution = resolution
        self.max_hourly_water = max_hourly_water
        self.sectors = zeros((self.resolution*self.height,
         self.resolution*self.width), Float)

    def apply_water(self, sprinkler):
        # adjusting coordinates to array indexes
        x, y = sprinkler.coords()
        ix = self.resolution * x
        iy = self.resolution * y
        ir = self.resolution * sprinkler.dispersion_radius
        for dx in range(int(ix-ir), int(ix+ir)):
            for dy in range(int(iy-ir), int(iy+ir)):
                if 0 <= dx and dx <= self.sectors.shape[1]-1 \
                 and 0 <= dy and dy <= self.sectors.shape[0]-1:
                    dc = sqrt(float(dx-ix)**2 + float(dy-iy)**2)
                    if dc <= ir:
                        try:
                            self.sectors[int(dy)][int(dx)] =  \
                             self.sectors[int(dy)][int(dx)] + \
                             sprinkler.dispersion(dc/self.resolution, 
                              out_units="centimeters_per_hour")
                        except IndexError:
                            raise str((dy,dx))

    def clear(self):
        self.sectors = zeros(self.sectors.shape, Float)

    def render_field(self, sprinkler_coords, path = "irrigation.png"):
        print "Drawing Image..."
        sectors = self.sectors * 100.0
        for x,y in sprinkler_coords:
            x = int(x * self.resolution)
            y = int(y * self.resolution)
            for ix in range(x-10, x+10):
                for iy in range(y-10, y+10):
                    if 0 <= iy and iy <= sectors.shape[0]-1 and \
                     0 <= ix and ix <= sectors.shape[1]-1:
                        try:
                            sectors[iy][ix] = 0.0
                        except:
                            raise str(((iy,ix), sectors.shape))
        im = Image.new("RGB", 
         (int(self.width*self.resolution), int(self.height*self.resolution)))
        img_array = zeros((self.width*self.resolution**2*self.height, 3), 'l')
        img_array[:,0] = ravel(sectors).astype("l")
        img_array[:,1] = ravel(sectors).astype("l")
        img_array[:,2] = ravel(sectors).astype("l")
        im.putdata(tools.tupleize(img_array.tolist()))
        im.save(path)


class Simulation:

    resolution = 10.0 #  number of array elements per square meter
    max_hourly_water = 0.750 # cm

    def __init__(self, sprinkler_positions, theta):
        self.f = Field(self.resolution, self.max_hourly_water)
        self.hourly = Field(self.resolution, self.max_hourly_water)
        self.s = SprinklerSystem(sprinkler_positions, theta)

    def run_simulation(self):
#        positions = [(0, 5.0, pi/2),
#                    ( 0, 5.0, pi/2),
#                    ( 0, 5.0, pi/2),
#                    ( 0, 5.0, pi/2),
#                    (20, 5.0, pi/2),
#                    (20, 5.0, pi/2),
#                    (20, 5.0, pi/2),
#                    (20, 5.0, pi/2),
#                    (40, 5.0, pi/2),
#                    (40, 5.0, pi/2),
#                    (40, 5.0, pi/2),
#                    (40, 5.0, pi/2),
#                    (60, 5.0, pi/2),
#                    (60, 5.0, pi/2),
#                    (60, 5.0, pi/2),
#                    (60, 5.0, pi/2),
#                    (80, 5.0, pi/2),
#                    (80, 5.0, pi/2),
#                    (80, 5.0, pi/2),
#                    (80, 5.0, pi/2)]
        
        positions = [( 1.0,  5.0, 0.0),
                     ( 1.0,  5.0, 0.0),
                     ( 5.0,  5.0, 0.0),
                     (30.0,  5.0, 0.0),
                     (30.0,  5.0, 0.0),
                     (55.0,  5.0, 0.0),
                     (59.0,  5.0, 0.0),
                     (59.0,  5.0, 0.0),
                     (59.0, 25.0, 0.0),
                     (59.0, 25.0, 0.0),
                     (55.0, 25.0, 0.0),
                     (30.0, 25.0, 0.0),
                     (30.0, 25.0, 0.0),
                     ( 5.0, 25.0, 0.0),
                     ( 1.0, 25.0, 0.0),
                     ( 1.0, 25.0, 0.0)]
        
        print "Starting simulation."
        for hours in range(len(positions)):
            print "Hour: %i, position: %s" % (hours, str(positions[hours]))
            self.hourly.clear()
            self.s.move(*positions[hours])
            print "  Running sprinklers."
            sprinkler_coords = []
            for sprink in self.s.run_sprinklers():
                print "   ", sprink.coords()
                sprinkler_coords.append(sprink.coords())
                self.hourly.apply_water(sprink)

            print "Maximum hourly accumulation on a sector: %f" % max(max(self.hourly.sectors))
            if max(self.hourly.sectors) > self.max_hourly_water:
                print "Too much water per hour on some sectors..."
            self.f.sectors = self.f.sectors + self.hourly.sectors
            self.f.render_field(sprinkler_coords, "irrigation_%i.png" % hours)
        print "Most accumulation on a sector: %f" % max(max(self.f.sectors))
        print "Least accumulation on a sector: %f" % min(min(self.f.sectors))


if __name__ == "__main__":
    s = Simulation([0, 4, 16, 20], pi/4)
    s.run_simulation()
