G = 9.81 # m/s^2

import types
from math import *
from matplotlib.patches import Circle, Rectangle, Polygon
from matplotlib.transforms import Value
from matplotlib.backends.backend_agg import RendererAgg


# all conversions to meters and seconds
in_conv = {"liters_per_minute": 1.66666666666e-5, # cubic meters per second
           "cubic_meters_per_second": 1.0,
           "centimeters": 1.0e-2, # meters
           "square_meters": 1.0, # leave alone
           "radians": 1.0,
}

# all conversions from meters and seconds
out_conv = {"meters_per_second": 1.0,
            "centimeters_per_second": 1.0e2, # from meters per second
            "centimeters_per_hour": 1.0e2 * 60.0 * 60.0 # from meters per second
}
        
def units(f):
    def wrapper(*args, **kwds):
        if kwds.has_key("in_units"):
            result = f(*[in_conv[c] * a for c, a in zip(kwds["in_units"],args)])
        else:
            result = f(*args)
        if kwds.has_key("out_units"):
            result = out_conv[kwds["out_units"]] * result
        return result
    return wrapper

@units
def velocity(flowrate, diameter):
    """Velocity as a function of flowrate and tube diameter."""
    # flowrate is in liters per minute (10^3 cm^3 / minute)
    # diameter in centimeters
    # returns meters per second
    return (4 * flowrate)/(pi * (diameter)**2)

@units
def radius(velocity, theta):
    """Radius as a function of velocity and angle."""
    # velocity in meters per second
    # theta (angle from horizontal, radians)
    # returns meters
    v_u = velocity * cos(theta)
    tm = 2 * v_u / G
    return velocity * sin(theta) * tm

@units
def area(radius):
    return pi * radius**2

@units
def depth(sprinkler_flow, area):
    return sprinkler_flow/area

@units
def find_m(n, radius):
    ds = 20.0/n
    m = 2 * floor(radius/ds) + 1
    if m >= n:
        return n
    else:
        return m

def tupleize(alist):
    return list([tuple(item) for item in alist])

def print_params(n, theta):
    # generates a parameter table
    f_in = 150.0 # L/min
    d_in = 0.6 # cm 
    print " # _flow_ n_vel radius ___area___ _depth_ _day_ _4days_  m _m*d_"
    for ii in range(1, n+1):
        F = f_in/ii
        V = velocity(F, d_in, in_units=["liters_per_minute", "centimeters"])
        R = radius(V, theta)
        A = area(R)
        D = depth(F, A,
         in_units=["liters_per_minute", "square_meters"],
         out_units="centimeters_per_hour")
        M = find_m(ii, R)
        print "%2i %6.2f %5.2f %6.2f %10.2f %5.5f %5.2f %7.3f %2i %4.3f" \
         % (ii, F, V, R, A, D, 24*D, 4*24*D, M, M*D)



