/*
 * Irrigation.c - irrigation simulator
 */

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <Ilib.h>

const double pi = 3.14159265358979323846433832975;

struct lines {
	char *line;
	struct lines *next;
};

struct positions {
	int duration;
	int vertical;
	float x;
	float y;
	struct positions *next;
} positions;

uint32_t field[1600][600];

double theta = 3.14159265358979323846433832795/4;

int nozzles;

float nozzlepositions[30];

void parserules(char *stuff);
struct lines *parselines(char *stuff);
void simulate();
void spraywater(float x, float y, int duration);
void writeimage();

int main(int argc, char **argv)
{
	int i, j;
	
	if (argc < 2) {
		printf("Please enter a rules file\n");
		exit(-1);
	}

	parserules(argv[1]);

	for (i = 0; i < 1600; i++)
		for (j = 0; j < 600; j++)
			field[i][j] = 0;

	simulate();
	writeimage();

	return 0;
}

void writeimage()
{
	long long avg = 0;
	double variance = 0;
	int i, j, l = 255, m = 0, sum = 960000;
	IColor c;
	IGC g = ICreateGC();
	//IImage out = ICreateImage(1600, 600, IOPTION_GREYSCALE);
	IImage out = ICreateImage(1600, 600, 0);


	
	for (i = 0; i < 1600; i++)
		for (j = 0; j < 600; j++) {
			if (field[i][j] > 255) {
				field[i][j] = 255;
				sum--;
			} else {
				avg += field[i][j];
				if (field[i][j] > m)
					m = field[i][j];
			}
			if (field[i][j] < l)
				l = field[i][j];
			c = IAllocColor(255 - field[i][j], 255 - field[i][j], 255 - field[i][j]);
			//c = IAllocColor(0, 0, 255 - field[i][j]);
			ISetForeground(g, c);
			IDrawPoint(out, g, i, j);
		}
	printf("Greatest acculumation: %i\nSmallest accumulation: %i\n", m, l);
	float rat = (float) m / l;
	double a2 = (double) avg / sum;
	printf("Ratio: %f, Average: %f\n", rat, a2);

	for (i = 0; i < 1600; i++)
		for (j = 0; j < 600; j++)
			if (field[i][j] < 255)
				variance += (a2 - field[i][j]) * (a2 - field[i][j]);
	
	double v2 = variance / sum;
	printf("Standard deviation: %f, percent %f%%\n", sqrt(v2), ((sqrt(v2) / a2) * 100));

	FILE *fp = fopen("irrigation.png", "w");
	IWriteImageFile(fp, out, IFORMAT_PNG, 0);
}

void simulate()
{
	int i, n = 0;
	struct positions *p;

	p = positions.next;
	while (p) {
		for (i = 0; i < nozzles; i++) {
			if (p->vertical)
				spraywater(p->x, p->y + nozzlepositions[i], p->duration);
			else
				spraywater(p->x + nozzlepositions[i], p->y, p->duration);
		}
		
		p = p->next;
	}
}

void parserules(char *stuff)
{
	int i;
	struct stat statbuf;
	char *file, *p;
	struct lines *lines;
	struct positions *q = &positions;

	i = open(stuff, O_RDONLY);

	if (i < 0) {
		printf("Can't open rules file\n");
		exit(-1);
	}

	fstat(i, &statbuf);

	file = malloc(statbuf.st_size + 1);
	read(i, file, statbuf.st_size);
	close(i);
	i = 0;

	lines = parselines(file);

	while (lines) {
		if (strncmp(lines->line, "nozzle", 6) == 0) {
			lines->line += 6;
			for (nozzles = 0; ; nozzles++) {
				nozzlepositions[nozzles] = strtof(lines->line, &p);
				if (lines->line == p)
					break;
				lines->line = p;
			}
		} else if (strncmp(lines->line, "position", 8) == 0) {
			lines->line += 8;
			q = q->next = malloc(sizeof(struct positions));
			
			q->duration = strtol(lines->line, &p, 10);
			lines->line = p;
			while (isblank(*lines->line))
				lines->line++;
			
			q->vertical = 0;
			if (*lines->line == 'v' || *lines->line == 'V')
				q->vertical = 1;
			lines->line++;

			q->x = strtof(lines->line, &p);
			q->y = strtof(p, NULL);
		}
		i++;
		lines = lines->next;
	}
	q->next = NULL;
}

struct lines *parselines(char *stuff)
{
	struct lines l, *p = &l;

	while (1) {
		p = p->next = malloc(sizeof(struct lines));
		p->line = stuff;

		stuff = strchr(stuff, '\n');
		if (!stuff)
			break;
		*stuff = '\0';
		++stuff;
	}
	p->next = NULL;
	return l.next;
}

void spraywater(float x, float y, int duration)
{
	// decimeters, seconds
	int i, j, ix, iy, a, b;
	
	ix = x * 20;
	iy = y * 20;

	double flow = 2.5 * 1000 / nozzles;
	
	double velocity = (4 * flow)/(pi * .36 * 100);

	double vu = velocity * cos(theta);
	double tm = 2 * vu / 9.81;
	double radius = velocity * sin(theta) * tm;
	radius *= 20;
	radius *= .9;
	double r2 = radius * radius;

	for (i = 0; i < 1600; i++)
		for (j = 0; j < 600; j++) {
			 a = i - ix;
			 b = j - iy;
			 if (a * a + b * b < r2) {
				 if (a * a + b * b < 120)
					 field[i][j] = 256;
				 field[i][j] += (((radius - sqrt(a * a + b * b))) ) * flow *40/ r2;
				 //field[i][j] += (((sqrt(a * a + b * b))) ) * flow *10/ r2;
			 }
		}
}
